from typing import Tuple, Union, List
import numpy as np
from time import sleep

from bluer_options.logger import crash_report
from bluer_options import string
from bluer_objects import file, objects

from bluer_sbc.imager.camera.generic import Camera
from bluer_sbc.hardware import hardware
from bluer_sbc import env
from bluer_sbc.logger import logger


class RPI_Camera(Camera):
    def capture_function(self) -> Tuple[bool, np.ndarray]:
        success = False
        image = np.ones((1, 1, 3), dtype=np.uint8) * 127

        temp = file.auxiliary("camera", "png")
        try:
            self.device.capture(temp)
            success = True
        except Exception as e:
            crash_report(e)

        if success:
            success, image = file.load_image(temp)

        return success, image

    # https://projects.raspberrypi.org/en/projects/getting-started-with-picamera/6
    def capture_video(
        self,
        filename: str,
        object_name: str,
        length: int = 10,  # in seconds
        preview: bool = True,
        pulse: bool = True,
        resolution=None,
    ) -> bool:
        if not self.open(resolution=resolution):
            return False

        full_filename = objects.path_of(
            object_name=object_name,
            filename=filename,
        )

        success = True
        try:
            if preview:
                self.device.start_preview()

            self.device.start_recording(full_filename)
            if pulse:
                for _ in range(int(10 * length)):
                    hardware.pulse("outputs")
                    sleep(0.1)
            else:
                sleep(length)
            self.device.stop_recording()

            if preview:
                self.device.stop_preview()
        except Exception as e:
            crash_report(e)
            success = False

        if not self.close():
            return False

        if success:
            logger.info(
                "{}.capture_video(): {} -{}-> {}".format(
                    self.__class__.__name__,
                    string.pretty_duration(length),
                    string.pretty_bytes(file.size(full_filename)),
                    filename,
                )
            )

        return success

    def close_function(self):
        self.device.close()

    def get_resolution_function(self) -> Tuple[int, int]:
        return [value for value in self.device.resolution]

    def open_function(
        self,
        resolution=Union[List[int], None],
    ):
        from picamera import PiCamera

        self.device = PiCamera()
        self.device.rotation = env.BLUER_SBC_CAMERA_ROTATION

        # https://projects.raspberrypi.org/en/projects/getting-started-with-picamera/7
        self.device.resolution = (
            (
                (2592, 1944)
                if env.BLUER_SBC_CAMERA_HI_RES
                else (
                    env.BLUER_SBC_CAMERA_WIDTH,
                    env.BLUER_SBC_CAMERA_HEIGHT,
                )
            )
            if resolution is None
            else resolution
        )



import argparse

from blueness import module
from bluer_options import string
from bluer_objects.env import abcli_object_name

from bluer_sbc import NAME
from bluer_sbc.imager.camera import instance as camera
from bluer_sbc.logger import logger

NAME = module.name(__file__, NAME)


parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="capture | capture_video | preview",
)
parser.add_argument(
    "--object_name",
    type=str,
    default=abcli_object_name,
)
parser.add_argument(
    "--filename",
    type=str,
    default="",
)
parser.add_argument(
    "--length",
    type=int,
    default=0,
)
parser.add_argument(
    "--output_path",
    type=str,
    default="",
)
parser.add_argument(
    "--preview",
    type=int,
    default=1,
    help="0 | 1",
)
args = parser.parse_args()

success = False
if args.task == "capture":
    success, image = camera.capture(
        filename=args.filename if args.filename else f"{string.timestamp()}.png",
        object_name=args.object_name,
    )
elif args.task == "capture_video":
    success = camera.capture_video(
        filename=args.filename if args.filename else f"{string.timestamp()}.h264",
        object_name=args.object_name,
        length=args.length if args.length else 10,
        preview=args.preview,
        resolution=(728, 600),
    )
elif args.task == "preview":
    success = camera.preview(length=args.length if args.length else -1)

else:
    logger.error(f"-{NAME}: {args.task}: command not found.")

if not success:
    logger.error(f"-{NAME}: {args.task}: failed.")



import cv2
from PIL import Image

from bluer_sbc.hardware.screen import Screen


class Adafruit_Rgb_Matrix(Screen):
    def __init__(self):
        super(Adafruit_Rgb_Matrix, self).__init__()
        self.size = (32, 32)
        self.animated = True

        from rgbmatrix import RGBMatrix, RGBMatrixOptions

        options = RGBMatrixOptions()
        options.rows = 32
        options.chain_length = 1
        options.parallel = 1
        options.hardware_mapping = "adafruit-hat"  # adafruit-hat/regular

        self.matrix = RGBMatrix(options=options)

    def update_screen(self, image, session, header):
        image = cv2.resize(image, self.size)

        super().update_screen(image, session, header)

        self.matrix.SetImage(Image.fromarray(image).convert("RGB"))

        return self



NAME = "bluer_sbc"

ICON = "🌀"

DESCRIPTION = f"{ICON} AI for single board computers."

VERSION = "8.163.1"

REPO_NAME = "bluer-sbc"

MARQUEE = "https://github.com/kamangir/blue-bracket/blob/main/images/helmet-3.jpg"

ALIAS = "@sbc"


def fullname() -> str:
    return f"{NAME}-{VERSION}"



import copy

from bluer_sbc.hardware.hardware import Hardware


class Screen(Hardware):
    def __init__(self):
        super().__init__()
        self.size = None

        self.buffer = None

    def update_screen(self, image, session, header):
        if self.animated:
            self.buffer = copy.deepcopy(image)

        return super().update_screen(image, session, header)



from bluer_options.env import load_config, load_env, get_env

load_config(__name__)
load_env(__name__)

BLUER_SBC_CAMERA_HI_RES = get_env("BLUER_SBC_CAMERA_HI_RES", True)
BLUER_SBC_CAMERA_WIDTH = get_env("BLUER_SBC_CAMERA_WIDTH", 728)
BLUER_SBC_CAMERA_HEIGHT = get_env("BLUER_SBC_CAMERA_HEIGHT", 600)
BLUER_SBC_CAMERA_ROTATION = get_env("BLUER_SBC_CAMERA_ROTATION", 0)
BLUER_SBC_CAMERA_KEEP_OPEN = get_env("BLUER_SBC_CAMERA_KEEP_OPEN", 0)

BLUER_SBC_DISPLAY_FULLSCREEN = get_env("BLUER_SBC_DISPLAY_FULLSCREEN", True)

BLUER_SBC_HARDWARE_KIND = get_env("BLUER_SBC_HARDWARE_KIND")

BLUER_SBC_SESSION_PLUGIN = get_env("BLUER_SBC_SESSION_PLUGIN")

BLUER_SBC_SESSION_IMAGER = get_env("BLUER_SBC_SESSION_IMAGER")
BLUER_SBC_SESSION_IMAGER_DIFF = get_env("BLUER_SBC_SESSION_IMAGER_DIFF", 0.1)
BLUER_SBC_SESSION_IMAGER_ENABLED = get_env("BLUER_SBC_SESSION_IMAGER_ENABLED", True)

BLUER_SBC_SESSION_MONITOR_ENABLED = get_env("BLUER_SBC_SESSION_MONITOR_ENABLED", True)

BLUER_SBC_SESSION_OBJECT_TAGS = get_env("BLUER_SBC_SESSION_OBJECT_TAGS")

BLUER_SBC_SESSION_OUTBOUND_QUEUE = get_env("BLUER_SBC_SESSION_OUTBOUND_QUEUE")

BLUER_SBC_SESSION_AUTO_UPLOAD = get_env("BLUER_SBC_SESSION_AUTO_UPLOAD", True)

BLUER_SBC_SESSION_IMAGER_PERIOD = get_env("BLUER_SBC_SESSION_IMAGER_PERIOD", 300)
BLUER_SBC_SESSION_MESSENGER_PERIOD = get_env("BLUER_SBC_SESSION_MESSENGER_PERIOD", 60)
BLUER_SBC_SESSION_REBOOT_PERIOD = get_env("BLUER_SBC_SESSION_REBOOT_PERIOD", 14400)
BLUER_SBC_SESSION_SCREEN_PERIOD = get_env("BLUER_SBC_SESSION_SCREEN_PERIOD", 4)
BLUER_SBC_SESSION_TEMPERATURE_PERIOD = get_env(
    "BLUER_SBC_SESSION_TEMPERATURE_PERIOD", 300
)

BLUER_SBC_ENV = get_env("BLUER_SBC_ENV")



from bluer_objects import README

image_template = "https://github.com/kamangir/assets2/blob/main/cheshmak/{}?raw=true"

marquee = README.Items(
    [
        {
            "name": "cheshmak",
            "marquee": image_template.format("01.png"),
            "url": "./bluer_sbc/docs/cheshmak.md",
        }
    ]
)

items = README.Items(
    [
        {
            "marquee": image_template.format(f"{index+1:02}.png"),
            "name": "",
        }
        for index in range(1)
    ]
)



import os
from typing import List

from bluer_ai.env import ABCLI_PATH_IGNORE
from bluer_objects import file

from bluer_sbc.logger import logger


def reply_to_bash(
    status: str,  # exit/reboot/seed/shutdown/update
    content: List[str] = [],
) -> bool:
    logger.info(f"session.reply_to_bash({status}).")

    return file.save_text(
        filename=os.path.join(
            ABCLI_PATH_IGNORE,
            f"session_reply_{status}",
        ),
        text=content,
    )



from blueness import module
from bluer_options import host

from bluer_sbc import NAME
from bluer_sbc.env import BLUER_SBC_HARDWARE_KIND
from bluer_sbc.hardware.hardware import Hardware as Hardware_Class
from bluer_sbc.logger import logger

NAME = module.name(__file__, NAME)


if host.is_mac():
    from bluer_sbc.hardware.display import Display as Hardware_Class
elif BLUER_SBC_HARDWARE_KIND == "adafruit_rgb_matrix":
    from bluer_sbc.hardware.adafruit_rgb_matrix import (
        Adafruit_Rgb_Matrix as Hardware_Class,
    )
elif BLUER_SBC_HARDWARE_KIND == "grove":
    from bluer_sbc.hardware.grove import Grove as Hardware_Class
elif BLUER_SBC_HARDWARE_KIND == "prototype_hat":
    if host.is_headless():
        from bluer_sbc.hardware.hat.prototype import Prototype_Hat as Hardware_Class
    else:
        from bluer_sbc.hardware.display import Display as Hardware_Class
elif BLUER_SBC_HARDWARE_KIND == "scroll_phat_hd":
    from bluer_sbc.hardware.scroll_phat_hd import Scroll_Phat_HD as Hardware_Class
elif BLUER_SBC_HARDWARE_KIND == "sparkfun-top-phat":
    from bluer_sbc.hardware.sparkfun_top_phat.classes import (
        Sparkfun_Top_phat as Hardware_Class,
    )
elif BLUER_SBC_HARDWARE_KIND == "unicorn_16x16":
    from bluer_sbc.hardware.unicorn_16x16 import Unicorn_16x16 as Hardware_Class

hardware = Hardware_Class()

logger.info(f"{NAME}: {BLUER_SBC_HARDWARE_KIND}: {hardware.__class__.__name__}")



from typing import List

from bluer_options.terminal import show_usage


def help_input(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        [
            "@sbc",
            "hat",
            "input",
        ],
        "read hat inputs.",
        mono=mono,
    )


def help_output(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        [
            "@sbc",
            "hat",
            "output",
            "<10101010>",
        ],
        "activate hat outputs.",
        mono=mono,
    )


def help_validate(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        [
            "@sbc",
            "hat",
            "validate",
        ],
        "validate hat.",
        mono=mono,
    )


help_functions = {
    "input": help_input,
    "output": help_output,
    "validate": help_validate,
}



import cv2
import time
from matplotlib import cm

from blueness import module

from bluer_sbc import NAME
from bluer_sbc.hardware.screen import Screen
from bluer_sbc.logger import logger

NAME = module.name(__file__, NAME)


class Sparkfun_Top_phat(Screen):
    def __init__(self):
        super(Sparkfun_Top_phat, self).__init__()
        self.size = (7, 17)
        self.animated = False

        import board
        import neopixel

        # https://learn.sparkfun.com/tutorials/sparkfun-top-phat-hookup-guide/ws2812b-leds
        self.pixel_count = 6
        self.pixels = neopixel.NeoPixel(
            board.D12,
            self.pixel_count,
            auto_write=False,
        )

        self.intensity = 48

        # https://matplotlib.org/stable/tutorials/colors/colormaps.html
        self.colormap = cm.get_cmap("copper", self.pixel_count)(range(self.pixel_count))

        self.pulse_cycle = 0

        # https://github.com/sparkfun/Top_pHAT_Button_Py/blob/main/examples/top_phat_button_ex1.py
        import top_phat_button

        self.buttons = top_phat_button.ToppHATButton()
        logger.info(f"{NAME}.connection status: {self.buttons.is_connected()}")

        self.buttons.pressed_interrupt_enable = False
        self.buttons.clicked_interrupt_enable = False

    def clock(self):
        super().clock()

        self.buttons.button_pressed  # These functions must be called to update button variables to their latest setting
        self.buttons.button_clicked  # These functions must be called to update button variables to their latest setting

        if self.buttons.a_clicked == True:
            logger.info(f"{NAME}: a clicked: update.")
            self.key_buffer += ["u"]

        if self.buttons.b_clicked == True:
            logger.info(f"{NAME}: b clicked: shutdown.")
            self.key_buffer += ["s"]

        if self.buttons.center_clicked == True:
            logger.info(f"{NAME}: center clicked.")
            self.key_buffer += [" "]

        if self.buttons.up_clicked == True:
            logger.info(f"{NAME}: up clicked.")
            self.intensity = min(255, 2 * self.intensity)

        if self.buttons.down_clicked == True:
            logger.info(f"{NAME}: down clicked.")
            self.intensity = max(1, self.intensity // 2)

        return self

    def pulse(self, pin=None, frequency=None):
        super().pulse(pin, frequency)

        for index in range(self.pixel_count):
            self.pixels[index] = tuple(
                int(thing * self.intensity)
                for thing in self.colormap[
                    (index + int(self.pulse_cycle / 10)) % self.pixel_count
                ][:3]
            )

        self.pixels.show()

        self.pulse_cycle += 1

        return self

    def release(self):
        super().release()
        for index in range(self.pixel_count):
            self.pixels[index] = 3 * (0,)
            self.pixels.show()
            time.sleep(0.1)

    def update_screen(self, image, session, header):
        image = cv2.resize(image, self.size)

        super().update_screen(image, session, header)

        return self



import cv2

from bluer_sbc.hardware.screen import Screen


class Unicorn_16x16(Screen):
    def __init__(self):
        super().__init__()
        self.size = (16, 16)
        self.animated = True

    def release(self):
        super().release()

        import unicornhathd

        unicornhathd.off()

    def update_screen(self, image, session, header):
        import unicornhathd

        image = cv2.rotate(
            cv2.resize(
                image,
                self.size,
            ),
            cv2.ROTATE_90_CLOCKWISE,
        )

        super().update_screen(image, session, header)

        for x in range(0, 16):
            for y in range(0, 16):
                unicornhathd.set_pixel(
                    x,
                    y,
                    image[x, y, 0],
                    image[x, y, 1],
                    image[x, y, 2],
                )

        unicornhathd.show()

        return self



from typing import List

from bluer_options.terminal import show_usage, xtra


def help_validate(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "button | leds"

    return show_usage(
        [
            "@sbc",
            "sparkfun_top_phat",
            "validate",
            f"[{options}]",
        ],
        "validate sparkfun_top_phat.",
        mono=mono,
    )


help_functions = {
    "validate": help_validate,
}



import cv2
import numpy as np
import time

from blueness import module
from bluer_options import string
from bluer_options.logger import crash_report

from bluer_sbc import NAME
from bluer_sbc.logger import logger

NAME = module.name(__file__, NAME)


class Diff:
    def __init__(self, threshold=0.1):
        self.size = (60, 80)
        self.threshold = threshold

        self.last_diff = 0.0
        self.last_same_period = 0.0
        self.last_diff_time = time.time()

        self.previous = None

    def same(self, image):
        if self.threshold < 0:
            return False

        image_scaled = cv2.resize(image, self.size)

        if self.previous is None:
            self.previous = image_scaled
            self.last_diff_time = time.time()

            logger.info(
                f"{NAME}.diff.same({string.pretty_shape_of_matrix(image)}): initialized."
            )
            return False

        try:
            self.last_diff = float(
                np.percentile(
                    np.abs(image_scaled.astype(float) - self.previous.astype(float)),
                    90,
                )
                / 255
            )
            is_same = self.last_diff <= self.threshold

            logger.info(
                "{}.diff.same({}): {:.03f} - {}{}".format(
                    NAME,
                    string.pretty_shape_of_matrix(image),
                    self.last_diff,
                    ("!same,same".split(","))[int(is_same)],
                    " - {} since last diff.".format(
                        string.pretty_duration(
                            time.time() - self.last_diff_time,
                            include_ms=True,
                            largest=True,
                            short=True,
                        )
                    ),
                )
            )

            if not is_same:
                self.last_same_period = time.time() - self.last_diff_time
                self.last_diff_time = time.time()

            self.previous = image_scaled

            return is_same
        except Exception as e:
            crash_report(e)

        self.previous = None
        return False



from blueness import module

from bluer_sbc import NAME
from bluer_sbc import env
from bluer_sbc.logger import logger

NAME = module.name(__file__, NAME)


imager_name = env.BLUER_SBC_SESSION_IMAGER
if imager_name == "lepton":
    from bluer_sbc.imager.lepton import instance as imager
else:
    from bluer_sbc.imager.camera import instance as imager

logger.info(f"{NAME}: {imager_name}: {imager.__class__.__name__}")



import cv2
import time

from bluer_sbc.hardware.screen import Screen


class Scroll_Phat_HD(Screen):
    def __init__(self):
        super(Scroll_Phat_HD, self).__init__()
        self.size = (7, 17)
        self.animated = True

    def update_screen(self, image, session, header):
        import scrollphathd

        image = cv2.resize(
            image,
            self.size,
        )

        super().update_screen(image, session, header)

        image = cv2.cvtColor(
            image,
            cv2.COLOR_BGR2GRAY,
        )

        for y in range(0, 17):
            for x in range(0, 7):
                scrollphathd.set_pixel(y, x, image[y, x] / 255.0)

        time.sleep(0.01)
        scrollphathd.show()

        return self



import cv2
import numpy as np

from blueness import module
from bluer_options.logger import crash_report
from bluer_options.host import is_mac
from bluer_objects import file
from bluer_objects.graphics import add_signature
from bluer_objects.graphics.screen import get_size

from bluer_sbc import NAME
from bluer_sbc import env
from bluer_sbc.hardware.hat.prototype import Prototype_Hat
from bluer_sbc.host import signature
from bluer_sbc.logger import logger


NAME = module.name(__file__, NAME)


class Display(Prototype_Hat):
    def __init__(self):
        super().__init__()

        self.canvas = None
        self.canvas_size = (640, 480)

        self.title = " | ".join(signature())

        self.created = False

        self.sign_images = True
        self.interpolation = cv2.INTER_LINEAR

    def create(self):
        if self.created:
            return
        self.created = True

        logger.info(f"{NAME}.create()")

        if env.BLUER_SBC_DISPLAY_FULLSCREEN and not is_mac():
            # https://stackoverflow.com/a/34337534
            cv2.namedWindow(self.title, cv2.WND_PROP_FULLSCREEN)
            cv2.setWindowProperty(
                self.title, cv2.WND_PROP_FULLSCREEN, cv2.WINDOW_FULLSCREEN
            )

            screen_height, screen_width = get_size()
            self.canvas_size = (
                screen_width,
                screen_height,
            )
        else:
            cv2.namedWindow(self.title, cv2.WINDOW_NORMAL)
            cv2.resizeWindow(
                self.title,
                self.canvas_size[0],
                self.canvas_size[1],
            )

    def save(self, filename: str = "") -> str:
        if self.canvas is None:
            return ""

        if not filename:
            filename = file.auxiliary("display", "jpg")

        return filename if file.save_image(filename, self.canvas) else ""

    def update_gui(self):
        try:
            if len(self.canvas.shape) == 2:
                self.canvas = np.stack(3 * [self.canvas], axis=2)

            cv2.imshow(
                self.title,
                cv2.cvtColor(
                    cv2.resize(
                        self.canvas,
                        dsize=self.canvas_size,
                        interpolation=self.interpolation,
                    ),
                    cv2.COLOR_BGR2RGB,
                ),
            )
        except Exception as e:
            crash_report(e)

    def update_screen(self, image, session, header):
        super().update_screen(image, session, header)

        self.canvas = np.copy(image)

        if self.sign_images:
            self.canvas = add_signature(
                self.canvas,
                header=header,
                footer=[" | ".join(signature())],
            )

        self.create()

        self.update_gui()

        key = cv2.waitKey(1)
        if key not in [-1, 255]:
            key = chr(key).lower()
            logger.info(f"{NAME}.update_screen(): key: '{key}'")
            self.key_buffer.append(key)

        return self



from bluer_objects import README


image_template = (
    "https://github.com/kamangir/assets2/blob/main/bluer-swallow/design/{}?raw=true"
)

marquee = README.Items(
    [
        {
            "name": "bluer-swallow",
            "marquee": image_template.format("06.jpg"),
            "url": "./bluer_sbc/docs/bluer-swallow.md",
        }
    ]
)

items = README.Items(
    [
        {
            "marquee": image_template.format(f"{index+1:02}.jpg"),
            "name": "",
        }
        for index in range(6)
    ]
)
